/*********************************************************************************/
/* Optimized and multithreaded matrix product                                    */
/* S. Vialle - December 2019                                                     */
/*********************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <omp.h>
#include <cblas-openblas.h>

#include "kernel.h"
#include "init.h"


/*-------------------------------------------------------------------------------*/
/* Initialisation of local matrixes A, B and C                                   */
/*-------------------------------------------------------------------------------*/
void LocalMatrixInit(void)
{
 int i, j;                                /* Local matrix indexes                */

/* Initialization of the local matrix elements                                   */
 for (i = 0; i < SIZE; i++)
    for (j = 0; j < SIZE; j++)
       A[i][j] = (double) (i*SIZE + j);
 
 for (i = 0; i < SIZE; i++)
    for (j = 0; j < SIZE; j++) {
       B[i][j] = (double) (i*SIZE + j);
       TB[j][i] = B[i][j];
    }
 
 for (i = 0; i < SIZE; i++)
    for (j = 0; j < SIZE; j++)
       C[i][j] = 0.0;
}


/*-------------------------------------------------------------------------------*/
/* "Usage" function.                                                             */
/*-------------------------------------------------------------------------------*/
void usage(int ExitCode, FILE *std)
{
 fprintf(std,"MatrixProduct usage: \n");
 fprintf(std,"\t [-h]: print this help\n");
 fprintf(std,"\t [-k <Kernel Id> (default: %d)]\n",DEFAULT_KERNEL_ID);
 fprintf(std,"\t   0: naive kernel\n");
 fprintf(std,"\t   1: TB kernel\n");
 fprintf(std,"\t   2: TB + accumulator + big instruction kernel\n");
 fprintf(std,"\t   3: TB + accumulator array kernel\n");
 fprintf(std,"\t   4: ikj kernel\n");
 fprintf(std,"\t   5: BLAS dgemm kernel\n"); 
 fprintf(std,"\t [-nt <number of threads> (default: %d)]\n",DEFAULT_NB_THREADS);
 fprintf(std,"\t [-pl <id of the loop to parallelize> (default: %c)]\n",DEFAULT_PARLOOP_ID);
 fprintf(std,"\t   must be i, or j, or k\n");

 exit(ExitCode);
}


/*-------------------------------------------------------------------------------*/
/* Command Line parsing.                                                         */
/*-------------------------------------------------------------------------------*/
void CommandLineParsing(int argc, char *argv[])
{
 // Default init
 NbThreads = DEFAULT_NB_THREADS;
 KernelId = DEFAULT_KERNEL_ID;
 ParLoopId = DEFAULT_PARLOOP_ID;

 // Init from the command line
 argc--; argv++;
 while (argc > 0) {
     if (strcmp(argv[0],"-nt") == 0) {
       argc--; argv++;
       if (argc > 0) {
         NbThreads = atoi(argv[0]);
         argc--; argv++;
         if (NbThreads <= 0) {
           fprintf(stderr,"Error: number of thread has to be >= 1!\n");
           exit(EXIT_FAILURE);
         }
       } else {
         usage(EXIT_FAILURE, stderr);
       }
     } else if (strcmp(argv[0],"-k") == 0) {
       argc--; argv++;
       if (argc > 0) {
         KernelId = atoi(argv[0]);
         argc--; argv++;
         if (KernelId < 0 || KernelId >= NB_KERNELS) {
           fprintf(stderr,"Error: kernel Id has to be an integer in [0;%d]\n",NB_KERNELS - 1);
           exit(EXIT_FAILURE);
         }
       } else {
         usage(EXIT_FAILURE, stderr);
       }
     } else if (strcmp(argv[0],"-pl") == 0) {
       argc--; argv++;
       if (argc > 0 && (strlen(argv[0])) == 1) {
         ParLoopId = argv[0][0];
         argc--; argv++;
         if (ParLoopId < 'i' || ParLoopId > 'k') {
           fprintf(stderr,"Error: ParLoop Id has to be i, or j, or k\n");
           exit(EXIT_FAILURE);
         }
       } else {
         usage(EXIT_FAILURE, stderr);
       }
     } else if (strcmp(argv[0],"-h") == 0) {
       usage(EXIT_SUCCESS, stdout);
     } else {
       usage(EXIT_FAILURE, stderr);
     }
 }
 
 // Complementary inits
 openblas_set_num_threads(1);                    // Set OpenBLAS in sequential mode
}

