/*********************************************************************************/
/* Optimized and multithreaded matrix product                                    */
/* S. Vialle - December 2019                                                     */
/*********************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <omp.h>
#include <cblas-openblas.h>

#include "kernel.h"
#include "main.h"


/*-------------------------------------------------------------------------------*/
/* Global variable definitions.                                                  */
/*-------------------------------------------------------------------------------*/

/* Matrixes: C = A.B                                                             */
/* We use the Transposed B matrix, in place of B, to improve cache memory usage. */
double A[SIZE][SIZE];                            /* A Matrix.                    */
double B[SIZE][SIZE];                            /* B Matrix.                    */
double TB[SIZE][SIZE];                           /* Transposed B Matrix.         */
double C[SIZE][SIZE];                            /* C matrix (result matrix).    */

/* Global variables to control computations.                                     */
int NbThreads;
int KernelId;
char ParLoopId;


/*-------------------------------------------------------------------------------*/
/* Matrix products (different kernels)                                           */
/*-------------------------------------------------------------------------------*/
void LocalMatrixProduct()
{
  int i, j, k;

  // Set the number of OpenMP threads in the parallel regions
  omp_set_num_threads(NbThreads);

  switch (KernelId) {

  // Kernel 0: Naive code ----------------------------------------------------
  case 0 :
    for (i = 0; i < SIZE; i++) {
      for (j = 0; j < SIZE; j++) {
        for (k = 0; k < SIZE; k++) {
           C[i][j] += A[i][k] * B[k][j];
        }
      }
    }
    break;
    
  // Kernel 1: kernel based on A, C, and B transposed matrixes ---------------
  case 1:
    for (i = 0; i < SIZE; i++) {
      for (j = 0; j < SIZE; j++) {
        for (k = 0; k < SIZE; k++) {
           C[i][j] += A[i][k] * TB[j][k];
        }
      }
    }
    break;

  // Kernel 2: kernel based on A, C, TB, one accumulator and a big instruction
  case 2:
    for (i = 0; i < SIZE; i++) {
      for (j = 0; j < SIZE; j++) {
        double accu = 0.0;
        for (k = 0; k < (SIZE/8)*8; k += 8) {
           accu += A[i][k+0] * TB[j][k+0] +
                   A[i][k+1] * TB[j][k+1] +
                   A[i][k+2] * TB[j][k+2] +
                   A[i][k+3] * TB[j][k+3] +
                   A[i][k+4] * TB[j][k+4] +
                   A[i][k+5] * TB[j][k+5] +
                   A[i][k+6] * TB[j][k+6] +
                   A[i][k+7] * TB[j][k+7];
        }
        for (k = (SIZE/8)*8; k < SIZE; k++) {
           accu += A[i][k] * TB[j][k];
        } 
        C[i][j] = accu;
      }
    }
    break;

  // Kernel 3: kernel based on A, C, TB and on an accumulator array ----------
  case 3:
    switch (ParLoopId) {
    case 'i':
      // TO DO: Add parallelization of i loop 
      for (i = 0; i < SIZE; i++) {
        for (j = 0; j < SIZE; j++) {
          double accu[8] = {0.0};
          for (k = 0; k < (SIZE/8)*8; k += 8) {
             accu[0] += A[i][k+0] * TB[j][k+0];
             accu[1] += A[i][k+1] * TB[j][k+1];
             accu[2] += A[i][k+2] * TB[j][k+2];
             accu[3] += A[i][k+3] * TB[j][k+3];
             accu[4] += A[i][k+4] * TB[j][k+4];
             accu[5] += A[i][k+5] * TB[j][k+5];
             accu[6] += A[i][k+6] * TB[j][k+6];
             accu[7] += A[i][k+7] * TB[j][k+7];
          }
          for (k = (SIZE/8)*8; k < SIZE; k++) {
             accu[0] += A[i][k] * TB[j][k];
          } 
          C[i][j] = accu[0] + accu[1] + accu[2] + accu[3] + 
                    accu[4] + accu[5] + accu[6] + accu[7];
        }
      }
      break;
      
    case 'j':
      // TO DO: Add parallelization of j loop
      for (i = 0; i < SIZE; i++) {
        for (j = 0; j < SIZE; j++) {
          double accu[8] = {0.0};
          for (k = 0; k < (SIZE/8)*8; k += 8) {
            accu[0] += A[i][k+0] * TB[j][k+0];
            accu[1] += A[i][k+1] * TB[j][k+1];
            accu[2] += A[i][k+2] * TB[j][k+2];
            accu[3] += A[i][k+3] * TB[j][k+3];
            accu[4] += A[i][k+4] * TB[j][k+4];
            accu[5] += A[i][k+5] * TB[j][k+5];
            accu[6] += A[i][k+6] * TB[j][k+6];
            accu[7] += A[i][k+7] * TB[j][k+7];
          }
          for (k = (SIZE/8)*8; k < SIZE; k++) {
            accu[0] += A[i][k] * TB[j][k];
          } 
          C[i][j] = accu[0] + accu[1] + accu[2] + accu[3] + 
                    accu[4] + accu[5] + accu[6] + accu[7];
        }
      }
      break;
      
    case 'k':
      // TO DO: Add parallelization of k loop 
      for (i = 0; i < SIZE; i++) {
        for (j = 0; j < SIZE; j++) {
          double accu[8] = {0.0};
          for (k = 0; k < (SIZE/8)*8; k += 8) {
            accu[0] += A[i][k+0] * TB[j][k+0];
            accu[1] += A[i][k+1] * TB[j][k+1];
            accu[2] += A[i][k+2] * TB[j][k+2];
            accu[3] += A[i][k+3] * TB[j][k+3];
            accu[4] += A[i][k+4] * TB[j][k+4];
            accu[5] += A[i][k+5] * TB[j][k+5];
            accu[6] += A[i][k+6] * TB[j][k+6];
            accu[7] += A[i][k+7] * TB[j][k+7];
          }
          for (k = (SIZE/8)*8; k < SIZE; k++) {
            accu[0] += A[i][k] * TB[j][k];
          }
          C[i][j] = accu[0] + accu[1] + accu[2] + accu[3] + \
                    accu[4] + accu[5] + accu[6] + accu[7];
        }
      }
      break;
    } // end of switch on loop index in kernel 3
    
  break;

  // Kernel 4: kernel based on ikj nested loops ------------------------------
  case 4:
    switch (ParLoopId) {
    case 'i':
      // TO DO: Add parallelization of i loop
      for (i = 0; i < SIZE; i++) {
        for (k = 0; k < SIZE; k++) {
          for (j = 0; j < SIZE; j++) {
             C[i][j] += A[i][k] * B[k][j];
          }
        }
      }
      break;
      
    case 'j':
      // TO DO: Add parallelization of j loop  
      for (i = 0; i < SIZE; i++) {
        for (k = 0; k < SIZE; k++) {
          for (j = 0; j < SIZE; j++) {
            C[i][j] += A[i][k] * B[k][j];
          }
        }
      }
      break;
      
    case 'k':
      // TO DO: Add parallelization of k loop
      for (i = 0; i < SIZE; i++) {
        for (k = 0; k < SIZE; k++) {
          for (j = 0; j < SIZE; j++) {
            C[i][j] += A[i][k] * B[k][j];
          }
        }
      }
      break;
    } // end of switch on loop index in kernel 4
    
    break;

  // Kernel 5: BLAS dgemm routine --------------------------------------------
  case 5 :
    // TO DO: Add parallelization unsing OpenMP
    {
     int NbLines = SIZE; // Nb of lines of C to compute
     int FirstLine = 0;  // First line of A to read, and first line of C to write
     cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans,
                 NbLines, SIZE, SIZE,
                 1.0, &A[FirstLine][0], SIZE, 
                 &B[0][0], SIZE,
                 0.0, &C[FirstLine][0], SIZE);
    }
    break;

  //--------------------------------------------------------------------------
  default : 
    fprintf(stderr,"Error: kernel %d not implemented!\n",KernelId);
    exit(EXIT_FAILURE);
    break;
  }
}
